/**
 * @tanstack/router-core/src/index.ts
 *
 * Copyright (c) TanStack
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE.md file in the root directory of this source tree.
 *
 * @license MIT
 */
'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var history = require('@tanstack/history');
var invariant = require('tiny-invariant');
var tinyWarning = require('tiny-warning');
var path = require('./path.js');
var qss = require('./qss.js');
var route = require('./route.js');
var fileRoute = require('./fileRoute.js');
var router = require('./router.js');
var searchParams = require('./searchParams.js');
var utils = require('./utils.js');
var scrollRestoration = require('./scroll-restoration.js');
var defer = require('./defer.js');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var invariant__default = /*#__PURE__*/_interopDefaultLegacy(invariant);
var tinyWarning__default = /*#__PURE__*/_interopDefaultLegacy(tinyWarning);



Object.defineProperty(exports, 'invariant', {
	enumerable: true,
	get: function () { return invariant__default["default"]; }
});
Object.defineProperty(exports, 'warning', {
	enumerable: true,
	get: function () { return tinyWarning__default["default"]; }
});
exports.cleanPath = path.cleanPath;
exports.interpolatePath = path.interpolatePath;
exports.joinPaths = path.joinPaths;
exports.matchByPath = path.matchByPath;
exports.matchPathname = path.matchPathname;
exports.parsePathname = path.parsePathname;
exports.resolvePath = path.resolvePath;
exports.trimPath = path.trimPath;
exports.trimPathLeft = path.trimPathLeft;
exports.trimPathRight = path.trimPathRight;
exports.decode = qss.decode;
exports.encode = qss.encode;
exports.RootRoute = route.RootRoute;
exports.Route = route.Route;
exports.RouterContext = route.RouterContext;
exports.createRouteMask = route.createRouteMask;
exports.rootRouteId = route.rootRouteId;
exports.FileRoute = fileRoute.FileRoute;
exports.PathParamError = router.PathParamError;
exports.Router = router.Router;
exports.SearchParamError = router.SearchParamError;
exports.componentTypes = router.componentTypes;
exports.isMatchInvalid = router.isMatchInvalid;
exports.isRedirect = router.isRedirect;
exports.lazyFn = router.lazyFn;
exports.redirect = router.redirect;
exports.defaultParseSearch = searchParams.defaultParseSearch;
exports.defaultStringifySearch = searchParams.defaultStringifySearch;
exports.parseSearchWith = searchParams.parseSearchWith;
exports.stringifySearchWith = searchParams.stringifySearchWith;
exports.functionalUpdate = utils.functionalUpdate;
exports.isPlainObject = utils.isPlainObject;
exports.last = utils.last;
exports.partialDeepEqual = utils.partialDeepEqual;
exports.pick = utils.pick;
exports.replaceEqualDeep = utils.replaceEqualDeep;
exports.restoreScrollPositions = scrollRestoration.restoreScrollPositions;
exports.watchScrollPositions = scrollRestoration.watchScrollPositions;
exports.defer = defer.defer;
exports.isDehydratedDeferred = defer.isDehydratedDeferred;
Object.keys(history).forEach(function (k) {
	if (k !== 'default' && !exports.hasOwnProperty(k)) Object.defineProperty(exports, k, {
		enumerable: true,
		get: function () { return history[k]; }
	});
});
//# sourceMappingURL=index.js.map
